/*----------------------------------------------------------------------
  File    : nposs.c
  Contents: Naive possibilistic classifier management
  Author  : Christian Borgelt
  History : 2001.02.07 file created from file nbayes.c
            2001.02.09 first version of standalone module completed
            2001.07.15 parser improved (global variables removed)
            2001.07.17 adapted to modified module scan
            2004.08.12 adapted to new module parse (later removed)
            2007.02.14 adapted to modified module attset
            2011.12.12 adapted to modified scanner functions
            2013.08.09 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#ifndef TAB_OPC
#define TAB_OPC
#endif
#include "nposs.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* error codes 0 to -15 defined in scanner.h */
#define E_ATTEXP    (-16)       /* attribute expected */
#define E_UNKATT    (-17)       /* unknown attribute */
#define E_DUPATT    (-18)       /* duplicate attribute */
#define E_MISATT    (-19)       /* missing attribute */
#define E_INVATT    (-20)       /* invalid attribute */
#define E_CLSTYPE   (-21)       /* class attribute is not nominal */
#define E_CLSCNT    (-22)       /* class attribute has too few values */
#define E_CLSATT    (-23)       /* class attribute name expected */
#define E_CLSEXP    (-24)       /* class value expected */
#define E_UNKCLS    (-25)       /* unknown class value */
#define E_DUPCLS    (-26)       /* duplicate class value */
#define E_DUPICLS   (-27)       /* duplicate class value (implicit) */
#define E_MISCLS    (-28)       /* missing class value */
#define E_VALEXP    (-29)       /* attribute value expected */
#define E_UNKVAL    (-30)       /* unknown attribute value */
#define E_DUPVAL    (-31)       /* duplicate attribute value */
#define E_DUPIVAL   (-32)       /* duplicate attribute value (impl.) */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- selectable attribute --- */
  ATTID  attid;                 /* attribute identifier */
  double errs;                  /* number of misclassifications */
} SELATT;                       /* (selectable attribute) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef NPC_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_ATTEXP  -16 */  "#attribute expected instead of '%s'",
  /* E_UNKATT  -17 */  "#unknown attribute '%s'",
  /* E_DUPATT  -18 */  "#duplicate attribute '%s'",
  /* E_MISATT  -19 */  "#missing attribute '%s'",
  /* E_INVATT  -20 */  "#invalid attribute '%s'",
  /* E_CLSTYPE -21 */  "#class attribute '%s' is not nominal",
  /* E_CLSCNT  -22 */  "#class attribute '%s' has too few values",
  /* E_CLSATT  -23 */  "#class attribute '%s' expected instead of '%s'",
  /* E_CLSEXP  -24 */  "#class value expected instead of '%s'",
  /* E_UNKCLS  -25 */  "#unknown class value '%s'",
  /* E_DUPCLS  -26 */  "#duplicate class value '%s'",
  /* E_DUPICLS -27 */  "#duplicate class value '%s' (implicit)",
  /* E_MISCLS  -28 */  "#missing class value '%s'",
  /* E_VALEXP  -29 */  "#attribute value expected instead of '%s'",
  /* E_UNKVAL  -30 */  "#unknown attribute value '%s'",
  /* E_DUPVAL  -31 */  "#duplicate attribute value '%s'",
  /* E_DUPIVAL -32 */  "#duplicate attribute value '%s' (implicit)",
};
#endif

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/
#ifdef NPC_INDUCE

static int induce (NPC *npc, TABLE *table)
{                               /* --- induce a poss. classifier */
  ATTID  i;                     /* loop variables for attributes */
  VALID  k;                     /* loop variable  for values/classes */
  TPLID  n;                     /* loop variable  for tuples */
  ATTID  clsid;                 /* index of the class attribute */
  VALID  cls;                   /* value of the class attribute */
  ATTSET *sel;                  /* current selection of attributes */
  TABLE  *opc;                  /* table of one point coverages */
  TUPLE  *tpl;                  /* to traverse the tuples */
  CDIST  *dist;                 /* to traverse the distrib. vectors */
  POSSD  *possd;                /* to traverse the distributions */
  float  wgt;                   /* buffer for tuple weight */

  assert(npc && table);         /* check the function arguments */

  /* --- determine one point coverages --- */
  sel = as_create("opc", att_delete);
  if (!sel) return -1;          /* create a table for the */
  opc = tab_create("opc", sel, tpl_delete);  /* one point */
  if (!opc) { as_delete(sel); return -1; }   /* coverages */
  if (tab_colcopy(opc, table, TAB_MARKED) != 0) {
    tab_delete(opc, 1); return -1; }
  tab_reduce(opc);              /* determine one point coverages */
  if (tab_opc(opc, TAB_COND) != 0) {      /* for the subspace of */
    tab_delete(opc, 1); return -1; }      /* the sel. attributes */
  for (clsid = as_attcnt(sel); --clsid >= 0; )
    if (att_getmark(as_att(sel, clsid)) == npc->clsid)
      break;                    /* find the index of the class */
  assert(clsid >= 0);           /* attribute in the o.p.c. table */

  /* --- build the classifier --- */
  npc_clear(npc);               /* clear the classifier */
  npc->total = tab_getwgt(table, 0, TPLID_MAX);
  for (n = tab_tplcnt(opc); --n >= 0; ) {
    tpl = tab_tpl(opc, n);      /* traverse the one point coverages */
    cls = tpl_colval(tpl, clsid)->n;
    if (cls < 0) continue;      /* skip tuples with an unknown class */
    assert(cls < npc->clscnt);  /* check the class value */
    wgt = tpl_getwgt(tpl);      /* get the tuple weight */
    for (i = as_attcnt(sel); --i >= 0; ) {
      if (i == clsid) {         /* traverse the attributes */
        if (wgt > npc->frqs[cls]) npc->frqs[cls] = wgt;
        continue;               /* for the class attribute only */
      }                         /* update the maximum projection */
      dist  = npc->dists +att_getmark(as_att(sel, i));
      possd = dist->possds +cls;/* get the class spec. distribution */
      k = tpl_colval(tpl,i)->n; /* get the current attribute's value */
      assert(k < dist->cnt);    /* and check it */
      if (k >= 0) {             /* if the attribute value is known */
        if (wgt > possd->frqs[k]) possd->frqs[k] = wgt; }
      else {                    /* otherwise traverse all values */
        for (k = 0; k < dist->cnt; k++)
          if (wgt > possd->frqs[k]) possd->frqs[k] = wgt;
      }                         /* (update the maximum projection */
    }                           /* to the subspace scaffolded by */
  }                             /* the class and the attribute) */

  /* --- clean up --- */
  tab_delete(opc, 1);           /* delete the one point coverages */
  return 0;                     /* return 'ok' */
}  /* induce() */

/*--------------------------------------------------------------------*/

static int eval (NPC *npc, TABLE *table, int mode,
                 SELATT *selas, ATTID cnt)
{                               /* --- evaluate selectable attributes */
  ATTID i, k, m;                /* loop variables for attributes */
  TPLID n;                      /* loop variable  for tuples */
  VALID cls;                    /* predicted class */
  ATT   *att;                   /* to traverse the attributes */
  TUPLE *tpl;                   /* to traverse the tuples */

  assert(npc && table && selas  /* check the function arguments */
  &&    (cnt > 0) && (mode & (NPC_ADD|NPC_REMOVE)));
  for (i = 0; i < cnt; i++) {   /* traverse the selectable attributes */
    k = selas[i].attid;         /* get the attribute identifier */
    npc->dists[k].mark = m = (mode & NPC_ADD) ? k : -1;
    att = as_att(npc->attset, k);
    att_setmark(att, m);        /* mark/unmark the next attribute */
    if (induce(npc, table) != 0)
      return -1;                /* induce a possibilistic classifier */
    npc_setup(npc);             /* and set it up for execution */
    for (n = tab_tplcnt(table); --n >= 0; ) {
      tpl = tab_tpl(table, n);  /* traverse the original tuples */
      cls = tpl_colval(tpl, npc->clsid)->n;
      if (cls < 0) continue;    /* skip tuples with an unknown class */
      if (npc_exec(npc, tpl, NULL) != cls)
        selas[i].errs += tpl_getwgt(tpl);
    }                           /* count the number of misclassific. */
    npc->dists[k].mark = m = (mode & NPC_ADD) ? -1 : k;
    att_setmark(att, m);        /* restore the attribute mark */
  }                             /* (restore old set of attributes) */
  return 0;                     /* return 'ok' */
}  /* eval() */

#endif
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

NPC* npc_create (ATTSET *attset, ATTID clsid)
{                               /* --- create a naive poss. class. */
  ATTID i, n;                   /* loop variables for attributes */
  VALID k;                      /* loop variable  for values */
  NPC   *npc;                   /* created possibilistic classifier */
  ATT   *att;                   /* to traverse the attributes */
  CDIST *dist;                  /* to traverse the cond. distribs. */
  POSSD *possd;                 /* to traverse the poss. distribs. */

  assert(attset && (clsid >= 0) /* check the function arguments */
  &&    (clsid < as_attcnt(attset))
  &&    (att_type(as_att(attset, clsid)) == AT_NOM));

  /* --- create the classifier body --- */
  n   = as_attcnt(attset);      /* get the number of attributes */
  npc = (NPC*)malloc(sizeof(NPC) +(size_t)(n-1) *sizeof(CDIST));
  if (!npc) return NULL;        /* allocate the classifier body */
  for (i = 0; i < n; i++)       /* clear the distribution vectors */
    npc->dists[i].possds = NULL;/* (for a proper clean up on error) */
  npc->attset = attset;         /* and initialize the other fields */
  npc->attcnt = n;
  npc->clsatt = as_att(attset, npc->clsid = clsid);
  npc->clscnt = att_valcnt(as_att(attset, clsid));
  npc->total  = 0;

  /* --- initialize the class distributions --- */
  assert(npc->clscnt > 0);      /* check the number of classes */
  npc->frqs = (double*)malloc((size_t)npc->clscnt *4 *sizeof(double));
  if (!npc->frqs) { npc_delete(npc, 0); return NULL; }
  npc->priors = npc->frqs   +npc->clscnt;  /* create class arrays */
  npc->posts  = npc->priors +npc->clscnt;  /* and initialize them */
  npc->buf    = npc->posts  +npc->clscnt;
  memset(npc->frqs, 0, (size_t)npc->clscnt *sizeof(double));

  /* --- initialize the conditional distributions --- */
  for (i = 0; i < npc->attcnt; i++) {
    dist = npc->dists +i;       /* traverse all attributes */
    dist->mark = -1;            /* and unmark them */
    if (i == clsid) {           /* if this is the class attribute, */
      dist->type = 0; continue;}/* clear the type for easier recogn. */
    att = as_att(attset, i);    /* get the next attribute */
    dist->type = att_type(att); /* and its type */
    if (dist->type != AT_NOM)   /* skip all attributes */
      continue;                 /* that are not nominal */
    dist->cnt = att_valcnt(att);
    if (npc->clscnt <= 0)       /* if there are no classes, */
      continue;                 /* there is nothing else to do */
    dist->possds = (POSSD*)calloc((size_t)npc->clscnt, sizeof(POSSD));
    if (!dist->possds) { npc_delete(npc, 0); return NULL; }
    if (dist->cnt <= 0)         /* create an array of discrete dists. */
      continue;                 /* and check whether there are values */
    for (k = 0; k < npc->clscnt; k++) {
      possd = dist->possds +k;  /* traverse the poss. distributions */
      possd->frqs = (double*)malloc((size_t)dist->cnt
                                    *2 *sizeof(double));
      if (!possd->frqs) { npc_delete(npc, 0); return NULL; }
      possd->poss = possd->frqs +dist->cnt;
      memset(possd->frqs, 0, (size_t)dist->cnt *sizeof(double));
    }                           /* init. the value frequencies */
  }
  return npc;                   /* return the created classifier */
}  /* npc_create() */

/*--------------------------------------------------------------------*/

NPC* npc_clone (NPC *npc, int cloneas)
{                               /* --- clone a naive poss. class. */
  ATTID  i;                     /* loop variable for attributes */
  VALID  k;                     /* loop variable for values */
  NPC    *clone;                /* created classifier clone */
  ATTSET *attset;               /* clone of attribute set */
  CDIST  *da; const CDIST *sa;  /* to traverse the cond. distribs. */
  POSSD  *dd; const POSSD *sd;  /* to traverse the poss. distribs. */

  assert(npc);                  /* check the function argument */

  /* --- copy the classifier body --- */
  attset = npc->attset;         /* get the attribute set */
  if (cloneas) {                /* if the corresp. flag is set, */
    attset = as_clone(attset);  /* clone the attribute set */
    if (!attset) return NULL;   /* of the original classifier, */
  }                             /* and then create a classifier */
  clone = (NPC*)malloc(sizeof(NPC)
                     +(size_t)(npc->attcnt-1) *sizeof(CDIST));
  if (!clone) { if (cloneas) as_delete(attset); return NULL; }
  for (i = 0; i < npc->attcnt; i++)
    clone->dists[i].possds = NULL;
  clone->attset = attset;       /* clear the distribution vectors */
  clone->attcnt = npc->attcnt;  /* (for a proper clean up on error) */
  clone->clsid  = npc->clsid;   /* and copy the other fields */
  clone->clscnt = npc->clscnt;
  clone->total  = npc->total;

  /* --- copy the class distributions --- */
  assert(npc->clscnt > 0);      /* check the number of classes */
  clone->frqs = (double*)malloc((size_t)npc->clscnt *4 *sizeof(double));
  if (!clone->frqs) { npc_delete(clone, cloneas); return NULL; }
  clone->priors = clone->frqs   +npc->clscnt; /* create class arrays */
  clone->posts  = clone->priors +npc->clscnt; /* and initialize them */
  clone->buf    = clone->posts  +npc->clscnt;
  memcpy(clone->frqs, npc->frqs,(size_t)npc->clscnt *2 *sizeof(double));

  /* --- copy the conditional distributions --- */
  for (i = 0; i < npc->attcnt; i++) {
    sa = npc->dists   +i;       /* traverse the distribution arrays */
    da = clone->dists +i;       /* of source and clone classifier */
    da->mark = sa->mark;        /* copy the attribute mark, */
    da->type = sa->type;        /* the attribute type, */
    da->cnt  = sa->cnt;         /* the number of attribute values */
    if (sa->type != AT_NOM)     /* if the attribute is not nominal, */
      continue;                 /* there is nothing else to do */
    da->possds = (POSSD*)calloc((size_t)clone->clscnt, sizeof(POSSD));
    if (!da->possds) { npc_delete(clone, cloneas); return NULL; }
    if (sa->cnt <= 0)           /* create a array of discrete dists. */
      continue;                 /* check whether there are values */
    for (k = 0; k < npc->clscnt; k++) {
      sd = sa->possds +k;       /* traverse the discrete distribs. */
      dd = da->possds +k;       /* of source and clone classifier */
      dd->frqs = (double*)malloc((size_t)da->cnt *2 *sizeof(double));
      if (!dd->frqs) { npc_delete(clone, cloneas); return NULL; }
      dd->poss = dd->frqs +da->cnt;
      memcpy(dd->frqs, sd->frqs, (size_t)sa->cnt *2 *sizeof(double));
    }                           /* copy the value frequencies */
  }
  return clone;                 /* return the created clone */
}  /* npc_clone() */

/*--------------------------------------------------------------------*/

void npc_delete (NPC *npc, int delas)
{                               /* --- delete a naive poss. class. */
  ATTID i;                      /* loop variable for attributes */
  VALID k;                      /* loop variable for values */
  CDIST *dist;                  /* to traverse the distrib. vectors */

  assert(npc);                  /* check the function argument */
  for (i = 0; i < npc->attcnt; i++) {
    dist = npc->dists +i;       /* traverse all nominal attributes */
    if (!dist->possds) continue;
    for (k = 0; k < npc->clscnt; k++)
      if (dist->possds[k].frqs) free(dist->possds[k].frqs);
    free(dist->possds);         /* delete all frequency vectors */
  }                             /* and the distribution vectors */
  if (npc->frqs) free(npc->frqs);
  if (delas)     as_delete(npc->attset);
  free(npc);                    /* delete the classifier body */
}  /* npc_delete() */

/*--------------------------------------------------------------------*/

void npc_clear (NPC *npc)
{                               /* --- clear a naive poss. classifier */
  ATTID i;                      /* loop variable for attributes */
  VALID k;                      /* loop variable for values */
  CDIST *dist;                  /* to traverse the distrib. vectors */

  assert(npc);                  /* check the function argument */
  npc->total = npc->max = 0;    /* clear the total number of cases */
  memset(npc->frqs, 0, (size_t)npc->clscnt *sizeof(double));
  for (i = 0; i < npc->attcnt; i++) {
    dist = npc->dists +i;       /* traverse all nominal attributes */
    if (!dist->possds) continue;
    for (k = 0; k < npc->clscnt; k++) {
      dist->possds[k].max = 0;  /* traverse the distributions */
      memset(dist->possds[k].frqs, 0, (size_t)dist->cnt*sizeof(double));
    }                           /* clear the value frequencies */
  }                             /* and their maximum */
}  /* npc_clear() */

/*--------------------------------------------------------------------*/
#ifdef NPC_INDUCE

NPC* npc_induce (TABLE *table, ATTID clsid, int mode)
{                               /* --- induce a naive poss. class. */
  int    r = 0;                 /* result of function call */
  ATTID  i, k;                  /* loop variables for attributes */
  VALID  c;                     /* loop variable  for values/classes */
  TPLID  n;                     /* loop variable  for tuples */
  ATTID  cnt;                   /* number of selectable attributes */
  VALID  cls;                   /* predicted class */
  NPC    *npc;                  /* created classifier */
  ATTSET *attset;               /* attribute set of the classifier */
  SELATT *selas;                /* vector of selectable attributes */
  TUPLE  *tpl;                  /* to traverse the tuples */
  CDIST  *dist;                 /* to traverse the distrib. vectors */
  double *p;                    /* to traverse the class possib. */
  double max;                   /* maximum of class possibilities */
  double errs;                  /* weight sum of misclassified tuples */

  assert(table                  /* check the function arguments */
  &&    (clsid >= 0) && (clsid < tab_colcnt(table))
  &&    (att_type(as_att(tab_attset(table), clsid)) == AT_NOM));

  /* --- create a classifier --- */
  attset = tab_attset(table);   /* get the attribute set of the table */
  if (mode & NPC_CLONE) {       /* if the corresp. flag is set, */
    attset = as_clone(attset);  /* clone the attribute set */
    if (!attset) return NULL;   /* of the given data table, */
  }                             /* then create a classifier */
  npc = npc_create(attset, clsid);
  if (!npc) { if (mode & NPC_CLONE) as_delete(attset); return NULL; }

  /* --- build initial classifier --- */
  if (mode & NPC_ADD) {         /* if to add attributes */
    for (n = tab_tplcnt(table); --n >= 0; ) {
      tpl = tab_tpl(table, n);  /* traverse the tuples in the table */
      cls = tpl_colval(tpl, npc->clsid)->n;
      if (cls < 0) continue;    /* skip tuples with an unknown class */
      assert(cls < npc->clscnt);/* and check the class value */
      npc->frqs[cls] += tpl_getwgt(tpl);
    } }                         /* determine the class frequencies */
  else {                        /* if to remove attributes */
    for (i = 0; i < npc->attcnt; i++) {
      dist = npc->dists +i;     /* traverse all attributes */
      dist->mark = ((dist->type == 0) || (dist->type == AT_NOM)) ? i:-1;
      att_setmark(as_att(attset, i), dist->mark);
    }                           /* mark class and nominal attributes */
    if (induce(npc, table) != 0) {
      npc_delete(npc, mode & NPC_CLONE); return NULL; }
  }                             /* create a full classifier */
  npc_setup(npc);               /* set up classifier for execution */
  if (!(mode & (NPC_ADD|NPC_REMOVE))) /* if no simp. is requested, */
    return npc;                 /* return the created classifier */

  /* --- evaluate initial classifier --- */
  if (mode & NPC_ADD) {         /* if to add attributes */
    p = npc->frqs; errs = max = *p;
    for (c = npc->clscnt; --c > 0; ) {
      errs += *++p;             /* traverse the class frequencies, */
      if (*p > max) max = *p;   /* sum them, and determine their */
    }                           /* maximum (find the majority class) */
    errs -= max; }              /* compute the number of prior errors */
  else {                        /* if to remove attributes */
    for (errs = 0, n = tab_tplcnt(table); --n >= 0; ) {
      tpl = tab_tpl(table, n);  /* traverse the tuples in the table */
      cls = tpl_colval(tpl, clsid)->n;
      if (cls < 0) continue;    /* skip tuples with an unknown class */
      if (npc_exec(npc, tpl, NULL) != cls)
        errs += tpl_getwgt(tpl);/* classify the current tuple */
    }                           /* and determine the number */
  }                             /* of misclassifications */
  #ifndef NDEBUG
  fprintf(stderr, "\n%8g errors initially\n", errs);
  #endif                        /* print a counter for debugging */

  /* --- collect selectable attributes --- */
  selas = malloc((size_t)npc->attcnt *sizeof(SELATT));
  if (!selas) { npc_delete(npc, mode & NPC_CLONE); return NULL; }
  for (cnt = i = 0; i < npc->attcnt; i++) {
    dist = npc->dists +i;       /* traverse all attributes */
    dist->mark = -1;            /* and unmark them (not selected) */
    if ( (dist->type == 0)      /* except the class attribute */
    ||   (dist->type != AT_NOM) /* and all numerical attributes */
    ||  ((dist->type == AT_NOM) /* and all nominal attributes */
    &&   (dist->cnt  <= 0)))    /* that do not have any values */
      continue;
    if (mode & NPC_REMOVE)      /* if to remove attributes, */
      dist->mark = i;           /* mark all selectable attributes */
    selas[cnt++].attid = i;     /* note selectable attributes */
  }                             /* afterwards mark the class */
  npc->dists[npc->clsid].mark = npc->clsid;
  for (i = 0; i < npc->attcnt; i++) /* transfer attribute markers */
    att_setmark(as_att(attset, i), npc->dists[i].mark);

  /* --- select attributes --- */
  while ((cnt  > 0)             /* while there are selectable atts. */
  &&     (errs > 0)) {          /* and the classifier is not perfect */
    for (i = 0; i < cnt; i++)   /* traverse the selectable attributes */
      selas[i].errs = 0;        /* and clear the numbers of errors */
    r = eval(npc, table, mode, selas, cnt);
    if (r < 0) break;           /* evaluate selectable attributes */
    for (i = k = 0; ++i < cnt;) /* traverse the selectable attributes */
      if (selas[i].errs < selas[k].errs)
        k = i;                  /* find least number of errors */
    if ( (selas[k].errs >  errs)/* if more tuples were misclassified */
    ||  ((selas[k].errs >= errs) && (mode & NPC_ADD)))
      break;                    /* abort the selection loop */
    errs = selas[k].errs;       /* note the new number of errors and */
    i    = selas[k].attid;      /* mark/unmark the selected attribute */
    npc->dists[i].mark = (mode & NPC_ADD) ? i : -1;
    att_setmark(as_att(attset, i), npc->dists[i].mark);
    #ifndef NDEBUG
    fprintf(stderr, "%8g errors with %s\n",
                    errs, att_name(as_att(attset, i)));
    #endif                      /* print a counter for debugging */
    for (--cnt; k < cnt; k++)   /* remove the selected attribute */
      selas[k].attid = selas[k+1].attid;
  }
  free(selas);                  /* delete the selectable atts. array */

  if ((r < 0)                   /* if an error occurred or the class. */
  ||  (induce(npc, table) != 0)) {      /* cannot be reinduced, abort */
    npc_delete(npc, mode & NPC_CLONE); return NULL; }
  npc_setup(npc);               /* set up classifier for execution */
  return npc;                   /* and return the created classifier */
}  /* npc_induce() */

/*--------------------------------------------------------------------*/

ATTID npc_mark (NPC *npc)
{                               /* --- mark selected attributes */
  ATTID m;                      /* attribute marker */
  ATTID i;                      /* loop variable for attributes */
  ATTID cnt;                    /* attribute counter */

  assert(npc);                  /* check the function argument */
  for (cnt = i = 0; i < npc->attcnt; i++) {
    if (npc->dists[i].mark < 0) m = -1;
    else               { cnt++; m =  1; }
    att_setmark(as_att(npc->attset, i), m);
  }                             /* transfer marker to attribute set */
  att_setmark(npc->clsatt, 0);  /* mark the class attribute */
  return cnt;                   /* return number of marked atts. */
}  /* npc_mark() */

#endif
/*--------------------------------------------------------------------*/

void npc_setup (NPC *npc)
{                               /* --- set up a naive poss. class. */
  ATTID  i;                     /* loop variable  for attributes */
  VALID  k, n;                  /* loop variables for values/classes */
  CDIST  *dist;                 /* to traverse the distrib. vectors */
  POSSD  *possd;                /* to traverse the poss. distribs. */
  double cnt;                   /* number of cases */

  assert(npc);                  /* check the function argument */
  cnt = npc->total;             /* get the total number of cases */

  /* --- compute class possibilities --- */
  if (cnt > 0)                  /* if the denominator is valid, */
    for (k = 0; k < npc->clscnt; k++)   /* estimate class poss. */
      npc->priors[k] = npc->frqs[k] /cnt;
  else                          /* otherwise clear all possibilities */
    memset(npc->priors, 0, (size_t)npc->clscnt *sizeof(double));

  /* --- compute conditional possibilities --- */
  for (i = 0; i < npc->attcnt; i++) {
    dist = npc->dists +i;       /* traverse all attributes */
    if ((dist->type != AT_NOM)  /* skip the class attribute */
    ||  (dist->cnt  <= 0)       /* and nominal attributes */
    ||  (dist->mark <  0))      /* without any values */
      continue;                 /* and unmarked attributes */
    for (k = 0; k < npc->clscnt; k++) {
      possd = dist->possds +k;  /* traverse the distributions */
      n     = dist->cnt;        /* get the number of att. values */
      if (cnt > 0) {            /* if the denominator is valid */
        while (--n >= 0)        /* estimate the possibilities */
          possd->poss[n] = possd->frqs[n] /cnt; }
      else                      /* otherwise clear all possibilities */
        memset(possd->poss, 0, (size_t)n *sizeof(double));
    }                           /* (compute the value possibilities */
  }                             /* from the one point coverages) */
}  /* npc_setup() */

/*--------------------------------------------------------------------*/

VALID npc_exec (const NPC *npc, const TUPLE *tpl, double *conf)
{                               /* --- execute a naive poss. class. */
  ATTID       i;                /* loop variable  for attributes */
  VALID       k, c;             /* loop variables for values/classes */
  const CDIST *dist;            /* to traverse the distrib. vectors */
  const INST  *inst;            /* to traverse the instances */
  double      poss;             /* buffer for cond. possibility */

  assert(npc);                  /* check the function argument */
  memcpy(npc->posts, npc->priors, (size_t)npc->clscnt *sizeof(double));
  for (i = 0; i < npc->attcnt; i++) {
    dist = npc->dists +i;       /* traverse all attributes */
    if ((dist->type != AT_NOM)  /* except the class attribute, */
    ||  (dist->mark < 0))       /* all non-nominal attributes, */
      continue;                 /* and all unmarked attributes */
    inst = (tpl)                /* get the attribute instantiation */
         ? tpl_colval(tpl, i)   /* from the tuple or the att. set */
         : att_inst(as_att(npc->attset, i));
    if ((inst->n < 0) || (inst->n >= dist->cnt))
      continue;                 /* skip unknown and invalid values */
    for (k = 0; k < npc->clscnt; k++) {
      poss = dist->possds[k].poss[inst->n];
      if (poss < npc->posts[k]) npc->posts[k] = poss;
    }                           /* determine the minimum */
  }                             /* of the degrees of possibility */
  c = 0;                        /* find the "most possible" class */
  for (k = 1; k < npc->clscnt; k++)
    if (npc->posts[k] > npc->posts[c]) c = k;
  if (conf)                     /* if requested, */
    *conf = npc->posts[c];      /* get the confidence value */
  return c;                     /* return the classification result */
}  /* npc_exec() */

/*--------------------------------------------------------------------*/

int npc_desc (const NPC *npc, FILE *file, int mode, int maxlen)
{                               /* --- describe a naive poss. class. */
  ATTID       n;                /* loop variable for attributes */
  VALID       i, k;             /* loop variables for values */
  int         pos, max;         /* current and maximal position */
  int         len, ind, l;      /* name length and indentation */
  ATT         *att;             /* to traverse the attributes */
  const CDIST *dist;            /* to traverse the distrib. vectors */
  const POSSD *possd;           /* to traverse the poss. distribs. */
  char  buf[4*AS_MAXLEN+4];     /* output buffer for names */
  char  num[80];                /* output buffer for numbers */

  assert(npc && file);          /* check the function arguments */
  len = (maxlen > 0) ? maxlen -2 : 70;

  /* --- print a header (as a comment) --- */
  if (mode & NPC_TITLE) {       /* if the title flag is set */
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  naive possibilistic classifier\n");
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */
  max = (maxlen <= 0) ? INT_MAX : maxlen;

  /* --- start description --- */
  scn_format(buf, att_name(npc->clsatt), 0);
  fputs("npc(", file);          /* get the class attribute name */
  fputs(buf, file);             /* and print it */
  fputs(") = {\n", file);       /* start the classifier */
  fprintf(file, "  total = %g;\n", npc->total);

  /* --- print the class distribution --- */
  fputs("  poss(", file);       /* print a distribution indicator */
  fputs(buf, file);             /* print the class att. name and */
  fputs(") = {\n    ", file);   /* start the class distribution */
  pos = 4;                      /* initialize the output position */
  ind = att_valwd(npc->clsatt, 0) +2;  /* compute the indentation */
  for (i = 0; i < npc->clscnt; i++) {  /* traverse the classes */
    if (i > 0) fputs(",\n    ", file); /* print separator if nec. */
    len = (int)scn_format(buf, att_valname(npc->clsatt, i), 0);
    fputs(buf, file);           /* get and print the class name */
    for (pos = len+2; pos < ind; pos++)
      fputc(' ', file);         /* pad with blanks to equal width */
    fprintf(file, ": %g", npc->frqs[i]);
    if (mode & NPC_POSS)        /* print the absolute class frequency */
      fprintf(file, " (%.1f%%)", npc->priors[i] *100);
  }                             /* print the relative class frequency */
  fputs(" };\n", file);         /* terminate the class distribution */

  /* --- print the (conditional) distributions --- */
  for (n = 0; n < npc->attcnt; n++) {
    dist = npc->dists +n;       /* traverse all attributes */
    if ((dist->type != AT_NOM)  /* skip non-nominal attributes */
    ||  (dist->mark < 0))       /* and all unmarked attributes */
      continue;                 /* (including the class attribute) */
    fputs("  poss(", file);     /* print a distribution indicator */
    att = as_att(npc->attset,n);/* and get the next attribute */
    scn_format(buf, att_name(att), 0);
    fputs(buf, file);           /* print the attribute name */
    fputc('|', file);           /* and the condition separator */
    scn_format(buf, att_name(npc->clsatt), 0);
    fputs(buf, file);           /* print the class attribute name */
    fputs(") = {\n    ", file); /* and start the cond. distribution */
    for (i = 0; i < npc->clscnt; i++) {
      possd = dist->possds +i;  /* traverse the poss. distributions */
      if (i > 0)                /* if this is not the first class, */
        fputs(",\n    ", file); /* start a new output line */
      len = (int)scn_format(buf, att_valname(npc->clsatt, i), 0);
      fputs(buf, file);         /* get and print the class name */
      for (pos = len+2; pos < ind; pos++)
        fputc(' ', file);       /* pad with blanks to equal width */
      fputs(":{", file);        /* start the value distribution and */
      pos += 3;                 /* traverse the attribute values */
      for (k = 0; k < dist->cnt; k++) {
        if (k > 0) {            /* if this is not the first value, */
          fputc(',', file); pos++; }          /* print a separator */
        len  = (int)scn_format(buf, att_valname(att, k), 0);
        len += l = sprintf(num, ": %g", possd->frqs[k]);
        if (mode & NPC_POSS)    /* format value frequency */
          len += sprintf(num +l, " (%.1f%%)", possd->poss[k]*100);
        if ((pos     > ind)     /* if the line would get too long */
        &&  (pos+len > max-4)) {
          fputc('\n', file);    /* start a new line and indent */
          for (pos = 0; pos < ind; pos++) fputc(' ', file); }
        else {                  /* if there is enough space left, */
          fputc(' ', file); pos++; }    /* only print a separator */
        fputs(buf, file); fputs(num, file);
        pos += len;             /* print value and its frequency */
      }                         /* and update the output position */
      fputs(" }", file);        /* terminate the value distribution */
    }  /* for (i = 0; .. */
    fputs("};\n", file);        /* terminate the distributions */
  }  /* for (n = 0; .. */
  fputs("};\n", file);          /* terminate the classifier */

  /* --- print additional information (as a comment) --- */
  if (mode & NPC_INFO) {        /* if the add. info. flag is set */
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  number of attributes: %"ATTID_FMT"+1",
                  npc->attcnt-1);
    fprintf(file, "\n  number of tuples    : %g\n", npc->total);
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print additional classifier info. */
  }                             /* (as a comment) */
  return ferror(file) ? -1 : 0; /* return the write status */
}  /* npc_desc() */

/*--------------------------------------------------------------------*/
#ifdef NPC_PARSE

static int freqs (SCANNER *scan, int cls,
                  ATT *att, double *frqs, double *max)
{                               /* --- read a poss. distribution */
  VALID i, n;                   /* loop variable, number of values */
  int   t;                      /* buffer for token */

  assert(scan && att && frqs);  /* check the function arguments */
  SCN_CHAR(scan, '{');          /* consume '{' (start of distrib.) */
  for (i = n = att_valcnt(att); --i >= 0; )
    frqs[i] = -1;               /* clear the value frequencies */
  while (1) {                   /* attribute value read loop */
    t = scn_token(scan);        /* check for a name */
    if ((t != T_ID) && (t != T_NUM))
      SCN_ERRVAL(scan, (cls) ? E_CLSEXP : E_VALEXP);
    if (t != T_NUM) t = ':';    /* if the token is no number, */
    else {                      /* the token must be a value, */
      SCN_NEXT(scan);           /* otherwise consume the token, */
      t = scn_token(scan);      /* note the next token, and */
      scn_back(scan);           /* go back to the previous one */
    }                           /* (look ahead one token) */
    if (t == ':') {             /* if a  ':' follows */
      i = att_valid(att, scn_value(scan));
      if (i < 0)        SCN_ERRVAL(scan, (cls) ? E_UNKCLS : E_UNKVAL);
      if (frqs[i] >= 0) SCN_ERRVAL(scan, (cls) ? E_DUPCLS : E_DUPVAL);
      SCN_NEXT(scan);           /* get and consume the value */
      SCN_CHAR(scan, ':'); }    /* consume ':' */
    else {                      /* if no ':' follows */
      if (frqs[i = (i+1) % n] >= 0) {
        t = (cls) ? E_DUPICLS : E_DUPIVAL;
        SCN_ERROR(scan, t, att_valname(att, i));
      }                         /* get the cyclic successor id */
    }                           /* and check for a duplicate */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    frqs[i] = atof(scn_value(scan));
    if (frqs[i] < 0)              SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and check value frequency */
    if (scn_token(scan) == '('){/* if a relative number follows, */
      SCN_NEXT(scan);           /* consume '(' */
      if (scn_token(scan) != T_NUM)  SCN_ERROR(scan, E_NUMEXP);
      if (atof(scn_value(scan)) < 0) SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* consume the relative number */
      SCN_CHAR(scan, '%');      /* consume '%' */
      SCN_CHAR(scan, ')');      /* consume ')' */
    }
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop, */
  }                             /* otherwise consume ',' */
  SCN_CHAR(scan, '}');          /* consume '}' (end of distribution) */
  for (*max = 0, i = n; --i >= 0; ) {
    if      (frqs[i] < 0)    frqs[i] = 0;
    else if (frqs[i] > *max) *max = frqs[i];
  }                             /* clear the unset frequencies */
  return 0;                     /* return 'ok' */
}  /* freqs() */

/*--------------------------------------------------------------------*/

static int possds (NPC *npc, SCANNER *scan, ATT *att, POSSD *possds)
{                               /* --- read possibility distributions */
  VALID i;                      /* class identifier, buffer */
  int   t;                      /* buffer for token */

  assert(npc && scan && att && possds); /* check function arguments */
  for (i = npc->clscnt; --i >= 0; )
    possds[i].max = -1;         /* unmark all distributions */
  while (1) {                   /* distribution read loop */
    t = scn_token(scan);        /* check for a class name */
    if (t != '{') {             /* if a class name is given */
      if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_CLSEXP);
      i = att_valid(npc->clsatt, scn_value(scan));
      if (i < 0)                       SCN_ERRVAL(scan, E_UNKCLS);
      if (possds[i].max >= 0)          SCN_ERRVAL(scan, E_DUPCLS);
      SCN_NEXT(scan);           /* get and consume class value */
      SCN_CHAR(scan, ':'); }    /* consume ':' */
    else {                      /* if no class name is given */
      if (possds[i = (i+1) % npc->clscnt].max >= 0)
        SCN_ERROR(scan, E_DUPICLS, att_valname(npc->clsatt, i));
    }                           /* get and check cyclic successor id */
    possds[i].max = 0;          /* clear the maximum as a flag */
    t = freqs(scan, 0, att, possds[i].frqs, &possds[i].max);
    if (t) return t;            /* read the poss. distribution */
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop, */
  }                             /* otherwise consume ',' */
  for (i = 0; i < npc->clscnt; i++) {
    if ((npc->frqs[i] > 0) && (possds[i].max < 0))
      SCN_ERROR(scan, E_MISCLS, att_valname(npc->clsatt, i));
  }                             /* check for a complete classifier */
  return 0;                     /* return 'ok' */
}  /* possds() */

/*--------------------------------------------------------------------*/

static int conds (NPC *npc, SCANNER *scan)
{                               /* --- read cond. distributions */
  ATTID i;                      /* attribute identifier */
  int   t;                      /* buffer for token */
  CDIST *dist;                  /* to traverse the cond. distribs. */

  assert(npc && scan);          /* check the function arguments */
  while ((scn_token(scan) == T_ID) /* while another dist. follows */
  &&     ((strcmp(scn_value(scan), "poss") == 0)
  ||      (strcmp(scn_value(scan), "pi")   == 0)
  ||      (strcmp(scn_value(scan), "Pi")   == 0))) {
    SCN_NEXT(scan);             /* consume 'poss', 'pi' or 'Pi' */
    SCN_CHAR(scan, '(');        /* consume '(' */
    t = scn_token(scan);        /* check for a name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_ATTEXP);
    i = as_attid(npc->attset, scn_value(scan));
    if (i < 0)           SCN_ERRVAL(scan, E_UNKATT);
    dist = npc->dists +i;       /* get and check the attribute */
    if (dist->type == 0) SCN_ERRVAL(scan, E_INVATT);
    if (dist->mark >= 0) SCN_ERRVAL(scan, E_DUPATT);
    dist->mark = 1;             /* set the read flag */
    SCN_NEXT(scan);             /* consume the attribute name */
    SCN_CHAR(scan, '|');        /* consume '|' (condition indicator) */
    t = scn_token(scan);        /* get the next token */
    if (((t != T_ID) && (t != T_NUM))
    ||  (strcmp(scn_value(scan), att_name(npc->clsatt)) != 0))
      SCN_ERRVAL(scan, E_CLSEXP);
    SCN_NEXT(scan);             /* consume the class att. name */
    SCN_CHAR(scan, ')');        /* consume ')' */
    SCN_CHAR(scan, '=');        /* consume '=' */
    SCN_CHAR(scan, '{');        /* consume '{' */
    if (scn_token(scan) != '}'){/* if a distribution vector follows */
      t = possds(npc, scan, as_att(npc->attset, i), dist->possds);
      if (t) return t;          /* read conditional distributions */
    }                           /* and check for an error */
    SCN_CHAR(scan, '}');        /* consume '}' */
    SCN_CHAR(scan, ';');        /* consume ';' */
  }  /* while ((scn_token == T_ID) .. */
  return 0;                     /* return 'ok' */
}  /* conds() */

/*--------------------------------------------------------------------*/

static int parse (ATTSET *attset, SCANNER *scan, NPC **pnpc)
{                               /* --- parse naive poss. classifier */
  int   t, err = 0;             /* buffer for token, error flag */
  ATTID i;                      /* loop variable for attributes */
  ATT   *att;                   /* class attribute */
  NPC   *npc;                   /* created naive poss. classifier */

  /* --- read header --- */
  if ((scn_token(scan) != T_ID) /* check for 'npc' */
  || ((strcmp(scn_value(scan), "npc")        != 0)
  &&  (strcmp(scn_value(scan), "naive_poss") != 0)))
    SCN_ERROR(scan, E_STREXP, "npc");
  SCN_NEXT(scan);               /* consume 'npc' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  t = scn_token(scan);          /* check for an attribute name */
  if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_ATTEXP);
  i = as_attid(attset, scn_value(scan));
  if (i < 0)                       SCN_ERRVAL(scan, E_UNKATT);
  att = as_att(attset, i);      /* get and check the class attribute */
  if (att_type(att) != AT_NOM)     SCN_ERRVAL(scan, E_CLSTYPE);
  if (att_valcnt(att) < 1)         SCN_ERRVAL(scan, E_CLSCNT);
  *pnpc = npc = npc_create(attset, i);   /* create a naive */
  if (!npc) SCN_ERROR(scan, E_NOMEM);    /* poss. classifier */
  SCN_NEXT(scan);               /* consume the class name */
  SCN_CHAR(scan, ')');          /* consume '(' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */

  /* --- read total number of cases --- */
  if ((scn_token(scan) != T_ID) /* check for 'total' */
  ||  (strcmp(scn_value(scan), "total") != 0))
    SCN_ERROR(scan, E_STREXP, "total");
  SCN_NEXT(scan);               /* consume 'total' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_NUM) SCN_ERRVAL(scan, E_NUMEXP);
  npc->total = atof(scn_value(scan));
  if (npc->total < 0)           SCN_ERRVAL(scan, E_NUMBER);
  SCN_NEXT(scan);               /* consume the number of cases */
  SCN_CHAR(scan, ';');          /* consume ';' */

  /* --- read class distribution --- */
  if ((scn_token(scan) != T_ID) /* check for 'poss', 'pi' or 'Pi' */
  ||  ((strcmp(scn_value(scan), "poss") != 0)
  &&   (strcmp(scn_value(scan), "p")    != 0)
  &&   (strcmp(scn_value(scan), "P")    != 0)))
    SCN_ERROR(scan, E_STREXP, "poss");
  SCN_NEXT(scan);               /* consume 'poss', 'pi' or 'Pi' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  t = scn_token(scan);          /* get the next token */
  if (((t != T_ID) && (t != T_NUM))
  ||  (strcmp(scn_value(scan), att_name(att)) != 0))
    SCN_ERROR(scan, E_CLSATT, att_name(att), scn_value(scan));
  SCN_NEXT(scan);               /* consume the class att. name */
  SCN_CHAR(scan, ')');          /* consume ')' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  t = freqs(scan, 1, att, npc->frqs, &npc->max);
  if (t) return t;              /* read the class distribution */
  SCN_CHAR(scan, ';');          /* consume ';' */
  npc->dists[i].mark = 0;       /* mark the class attribute */

  /* --- read conditional distributions --- */
  do {                          /* read the distributions, */
    t = conds(npc, scan);       /* trying to recover on errors */
    if (t) { err = t; scn_recover(scan, ';', 0, 0, 0); }
  } while (t);                  /* while not all dists. read */
  if (err) return err;          /* if an error occurred, abort */
  for (i = 0; i < npc->attcnt; i++) {
    if ((npc->frqs[i] > 0) && (npc->dists[i].mark < 0))
      SCN_ERROR(scan, E_MISATT, att_name(as_att(attset, i)));
  }                             /* check for a complete classifier */

  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* parse() */

/*--------------------------------------------------------------------*/

NPC* npc_parse (ATTSET *attset, SCANNER *scan)
{                               /* --- parse a naive poss. class. */
  NPC *npc = NULL;              /* created naive poss. classifier */

  assert(attset);               /* check the function arguments */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (parse(attset, scan, &npc) != 0) {
    if (npc) npc_delete(npc,0); /* parse a naive poss. classifier */
    return NULL;                /* if an error occurred, */
  }                             /* delete the classifier and abort */
  npc_setup(npc);               /* set up the created classifier */
  return npc;                   /* and then return it */
}  /* npc_parse() */

#endif
